#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <ctype.h>
#include "utility.h"

#define USAGE "USAGE: markov_parameters seqfile\n"

#define MAX_NAME 128
#define MAX_LINE 2048

#define NUM_NUC 4

#define NUM_ROOT_PARAMETERS 48
#define NUM_TOTAL_ROOT_PARAMETERS 64

typedef struct {
  char* seq;
  char name[MAX_NAME];
} Sequence;

typedef struct {
  int n_seq;
  int seq_length;
  Sequence* sequence;
  double parameters[NUM_TOTAL_ROOT_PARAMETERS];
  int root_parameters[NUM_NUC][NUM_NUC][NUM_NUC];
  int root_enum[NUM_TOTAL_ROOT_PARAMETERS][3];
  int root_enum_sum[NUM_TOTAL_ROOT_PARAMETERS - NUM_ROOT_PARAMETERS][3];
  int root_enum_addend[NUM_ROOT_PARAMETERS];
} Params;

char NAUC[7] = "TCAGN-";
int complement[4] = {2,3,0,1};

void read_params(int argc, char* argv[], Params* params);
void read_sequence(Params* params, char* seqfile);
void print_parameters(Params* params);
void make_root_parameters(Params* params);


int main(int argc, char* argv[]) {
  Params *params;

  (params = (Params*)malloc(sizeof(Params))) || die("Could not allocate memory for params\n");

  read_params(argc, argv, params);
  print_parameters(params);
  return 0;
}

void print_parameters(Params* params) {
  int i;

  for(i = 0; i < NUM_TOTAL_ROOT_PARAMETERS; i++)
    printf("%c%c%c = %10.15g\n", NAUC[params->root_enum[i][0]], NAUC[params->root_enum[i][1]], NAUC[params->root_enum[i][2]], params->parameters[i]);
}

void read_params(int argc, char* argv[], Params* params) {
  int i;

  for(i = 1; i < argc && argv[i][0] == '-'; i++) {
    die(USAGE);
  }
  argc == i + 1 || die(USAGE);
  make_root_parameters(params);
  read_sequence(params, argv[i]);
}

void make_root_parameters(Params* params) {
  int i, j, k;
  int count = 0;

  for(i = 0; i < NUM_NUC; i++)
    for(k = 0; k < NUM_NUC; k++)
      for(j = 0; j < NUM_NUC; j++)
	params->root_parameters[i][j][k] = -1;
  
  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
      for(k = 0; k < NUM_NUC - 1; k++)
	if(params->root_parameters[i][j][k] == -1) {
	  params->root_parameters[i][j][k] = count;
	  params->root_enum[count][0] = i;
	  params->root_enum[count][1] = j;
	  params->root_enum[count][2] = k;
	  count++;
	}
  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
      if(params->root_parameters[i][j][NUM_NUC - 1] == -1) {
	params->root_parameters[i][j][NUM_NUC - 1] = count;
	params->root_enum[count][0] = i;
	params->root_enum[count][1] = j;
	params->root_enum[count][2] = NUM_NUC - 1;
	count++;
      }
  count == NUM_TOTAL_ROOT_PARAMETERS || die("miscounted root parameters: %d\n", count);

  for(i = NUM_ROOT_PARAMETERS; i < NUM_TOTAL_ROOT_PARAMETERS; i++)
    for(j = 0; j < 3; j++)
      params->root_enum_sum[i-NUM_ROOT_PARAMETERS][j] = params->root_parameters[params->root_enum[i][0]][params->root_enum[i][1]][(j < params->root_enum[i][2]) ? j : (j+1)];
  for(i = 0; i < NUM_ROOT_PARAMETERS; i++)
    params->root_enum_addend[i] = params->root_parameters[params->root_enum[i][0]][params->root_enum[i][1]][params->root_enum[i][0]];
}

void read_sequence(Params* params, char* seqfile) {
  FILE* in;
  int i, j, k, l, m, ch = EOF;
  char alpha_table[256];
  int dinuc[NUM_NUC][NUM_NUC];
  double trinuc[NUM_NUC][NUM_NUC][NUM_NUC];

  (in = fopen(seqfile, "r")) || die("Could not open seqfile: %s\n", seqfile);
  fscanf(in, "%d%d\n", &params->n_seq, &params->seq_length) == 2 || die("Error in seqfile: %s\n", seqfile);
  (params->sequence = (Sequence*)malloc(sizeof(Sequence) * params->n_seq)) || die("Could not allocate memory for params->sequence\n");
  for(i = 0; i < 256; i++)
    alpha_table[i] = -1;
  for(i = 0; i < 6; i++)
    alpha_table[NAUC[i]] = (char)i;
  for(i = 0; i < params->n_seq; i++) {
    fscanf(in, "%s.63", params->sequence[i].name) == 1 || die("Error in seqfile\n");
    (params->sequence[i].seq = (char*)malloc(sizeof(char) * params->seq_length)) || die("Could not allocate memory for params->sequence[%d].seq\n", i);
    for(j = 0; j < params->seq_length;) {
      if((ch = getc(in)) != EOF && alpha_table[ch] >= 0) {
	params->sequence[i].seq[j++] = alpha_table[ch];
      }
      else if(ch == '\n' || ch == EOF)
	die("Early sequence termination in sequence: %s, %d expected, %d found\n", params->sequence[i].name, params->seq_length, j + 1);
    }
    for(; ch != '\n' && ch != EOF; ch = getc(in));
  }
  fclose(in);

  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++) {
      dinuc[i][j] = 0;
      for(k = 0; k < NUM_NUC; k++)
        trinuc[i][j][k] = 0.0;
    }

  for(i = 0; i < params->n_seq; i++)
    for(j = 0; j < params->seq_length; j++)
     if(params->sequence[i].seq[j] < NUM_NUC && params->sequence[i].seq[(j+1) % params->seq_length] < NUM_NUC && params->sequence[i].seq[(j+2) % params->seq_length] < NUM_NUC) {
        trinuc[params->sequence[i].seq[j]][params->sequence[i].seq[(j+1) % params->seq_length]][params->sequence[i].seq[(j+2) % params->seq_length]]++;
	dinuc[params->sequence[i].seq[j]][params->sequence[i].seq[(j+1) % params->seq_length]]++;
      }

  for(i = 0; i < NUM_NUC; i++)
    for(j = 0; j < NUM_NUC; j++)
      for(k = 0; k < NUM_NUC; k++) {
	/*
	printf("%c%c->%c = %15.10g = %d / %d\n", NAUC[i], NAUC[j], NAUC[k], trinuc[i][j][k] / dinuc[i][j], (int)trinuc[i][j][k], dinuc[i][j]);
	*/
	params->parameters[params->root_parameters[i][j][k]] = trinuc[i][j][k] / dinuc[i][j];
      }
}

